#!/usr/bin/env python3
"""
    PKGval
    Verify PS3/PSP/Vita npdrm package checksums

    Copyright (C) 2022 hekel, pandorah.org

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
import os
import re
import hashlib


if sys.platform in ('win32', 'cygwin'):
    os.system('color')

class COL:
    GRN = "\033[38;5;41m"
    WARN = "\033[38;5;178m"
    FAIL = "\033[38;5;203m"
    CLR = "\033[0m"
    BOLD = "\033[1m"
    ULINE = "\033[4m"

def get_pkg_header(filename):
    with open(filename, 'rb') as pkg_file:
        header = pkg_file.read(232).hex().upper()

    magic_num = header[:8]
    pkg_rev = header[8:10]
    pkg_type = header[14:16]

    if magic_num == "7F504B47": # PKG # PS3/PSP/Vita
        contentid = bytes.fromhex(header[96:168]).decode()
        pkg_rev_type = {'80': "Retail", '00': "Debug"}
        if pkg_rev in pkg_rev_type:
            pkg_rev = pkg_rev_type[pkg_rev]
        if pkg_type == "01":
            pkg_type = "PS3"
        elif pkg_type == "02":
            pkg_type = "PSP/Vita"
            pkg_key_id = header[462:464]
            pkg_exthead_type = {'01': "PSP", '02': "PS Vita", '04': "PS Mobile"}
            if pkg_key_id in pkg_exthead_type:
                pkg_type = pkg_exthead_type[pkg_key_id]
        return {'type': pkg_type, 'rev': f"({pkg_rev})", 'contentid': contentid}

    elif magic_num == "7F434E54": # CNT # PS4
        contentid = bytes.fromhex(header[130:200]).decode()
        
        return {'type': "PS4", 'rev': f"({pkg_rev})", 'contentid': contentid}

def get_pkg_hash(filename, tail=None):
    BUF_SIZE = 65536  # 64KibiBytes
    sha1 = hashlib.sha1()

    file_size = os.stat(filename).st_size
    hashable_size = file_size
    if tail == None:
        hashable_size -= 32
    with open(filename, 'rb') as pkg_file:
        while hashable_size > 0:
            try:
                if BUF_SIZE > hashable_size:
                    BUF_SIZE = hashable_size
                print(f"Calculating SHA1: {100 - (100 * hashable_size/file_size):.0f}%", end='\r')
                chunk = pkg_file.read(BUF_SIZE)
                hashable_size -= BUF_SIZE
            except EOFError:
                break

            sha1.update(chunk)
            computed_sum = sha1.hexdigest().upper()
    return computed_sum

def get_pkg_npdrm_tail(filename):
    with open(filename, 'rb') as pkg_file:
        pkg_file.seek(-32, 2)
        npdrm_tail = pkg_file.read()
    return npdrm_tail.hex().upper()

def compare_hashes(hash_1, hash_2):
    if hash_1 == hash_2:
        print(f"{COL.GRN}Checksums match!{COL.CLR}")
    elif hash_1 != hash_2:
        print(f"{COL.FAIL}Checksums do not match!{COL.CLR}")    

def check_pkg(pkg_name):
    PKG = get_pkg_header(pkg_name)
    pkg_npdrm_tail = get_pkg_npdrm_tail(pkg_name)
    tail_padding = pkg_npdrm_tail[40:]
    pkg_tail_hash = pkg_npdrm_tail[:-24]
    digestible = ("PS3", "PSP/Vita", "PSP", "PS Vita", "PS Mobile")

    tail_hash_zeroed = re.search(r"^0+$", pkg_tail_hash)
    tail_pad_zeroed = re.search(r"^0+$", tail_padding)

    if PKG['type']:
        print(f"   Package Type : {PKG['type']} {PKG['rev']}")
        print(f"     Content ID : {PKG['contentid']}")
            
        if PKG['type'] in digestible:
            if not tail_hash_zeroed and tail_pad_zeroed:
                print(f" Package Digest : {pkg_tail_hash}")
            else:
                print(f"{COL.WARN}Notice:{COL.CLR} Package does not seem to contain an embedded Package Digest for comparison:")
                print(f"   Package Tail : {pkg_npdrm_tail}")
            pkg_sha1sum = get_pkg_hash(pkg_name)
        elif PKG['type'] == "PS4":
            pkg_sha1sum = get_pkg_hash(pkg_name, 'full')

        print(f"Calculated SHA1 : {pkg_sha1sum}")

    if PKG['type'] in digestible and not (tail_hash_zeroed and tail_pad_zeroed):
        compare_hashes(pkg_sha1sum, pkg_tail_hash)


usage = f"Usage:   {sys.argv[0]} FILENAME"
if len(sys.argv) > 1 and sys.argv[1] in ("--version", "-v"):
    print("PKGval v0.01 - Verify PlayStation npdrm package checksums\nCopyright \u00A9 2022 hekel, https://pandorah.org/pkgval\n")
    print("This program may be freely redistributed and/or modified under the terms\n\
of the GNU General Public License v3 or any later version as published\n\
by the Free Software Foundation. (https://www.gnu.org/licenses/gpl.html)")
else:
    try:
        check_pkg(sys.argv[1])
    except:
        print(usage)
